import unittest
import functools
from inspect import isgeneratorfunction
from typing import Callable, Dict, Iterator, Union, Optional

from flask import Flask
from flask.testing import FlaskClient
from flask.wrappers import Response


class LiveTestCase(unittest.TestCase):
    '''
    Interacts with a live flask webserver running on a daemon thread

    The server is started when the TestSuite (not TestCase) runs
    The lifetime of the server is the full duration of the TestSuite

    If you're testing flask using a headless browser (e.g selenium) - this is the testcase
    you need

    Should be used with LiveTestSuite
    '''
    app: Union[Flask, None] = None
    server_url: Union[str, None] = None


class _TestCaseImpl(unittest.TestCase):
    '''
    Common methods all the following TestCases will use

    This will set up methods to wrap up the `super().run` and `super().debug` call
    Similar set up is used by all the following testcases - with only slight variation
    The variation room is left open - to be implemented in `_handle_resource_instantiation_and_internal_call`
    by the actual testcases themselves

    The common part of the setup is-
    * Store the original setUp, testMethod (i.e the actual test defined by user),
      and tearDown methods in temp variables
    * Wrap the `super().run` and `super().debug` in a `try/finally` block
    * Restore the original setUp, testMethod and tearDown methods in the `finally` block
    '''

    ### Utility functions exposed to the user

    def assertStatus(self, rv: Response, expected_status: int):
        # Assert the status code of given response
        self.assertEqual(rv.status_code, expected_status)

    def assertResponseEqual(self, rv: Response, expected_response):
        # Assert the expected_response is equal to the given response's .data
        self.assertEqual(rv.data, expected_response)

    def assertJsonEqual(self, rv: Response, expected_json: Dict):
        # Assert the expected_json is equal to the given response's .json
        self.assertEqual(rv.json, expected_json)

    def assertInResponse(self, member, rv: Response):
        # Assert the given member exists in response.data
        self.assertIn(member, rv.data)

    def assertLocationHeader(self, rv: Response, expected_location: str):
        # Assert the expected_location matches the location key present in the given response's header
        self.assertEqual(rv.headers['Location'], expected_location)

    ### Override run and debug to provide the extra functionality

    def run(self, result: Optional[unittest.TestResult]) -> Optional[unittest.TestResult]:
        '''
        Wrap the actual `super().run` call with all the custom setup

        Prepare `super().run` with `result` and pass it to the setup
        This way, the setup can just call `internal_method()` without having to pass `result`
        '''
        return self._wrap_internal_method_with_custom_setup(functools.partial(super().run, result))

    def debug(self):
        '''
        Wrap the actual super().debug call with all the custom setup
        '''
        return self._wrap_internal_method_with_custom_setup(super().debug)

    ### Private methods that implement the extra functionality

    def _wrap_internal_method_with_custom_setup(self, internal_method: Callable):
        '''
        The test method currently being tested should be in _testMethodName
        The method will be overridden a bit, so the original should be stored safely

        The set up and tear down methods will also be overriden and should also be stored
        '''
        orig_test = getattr(self, self._testMethodName)
        orig_setup = self.setUp
        orig_teardown = self.tearDown
        '''
        This will manage the actual instantiation of resources (creating app/client or both)
        + call `super().run`/`super().debug`
        + clean up after itself
        '''
        return self._handle_resource_instantiation_and_internal_call(
            internal_method, orig_test, orig_setup, orig_teardown
        )

    def _handle_resource_instantiation_and_internal_call(
        self, internal_method: Callable, orig_test: Callable, orig_setup: Callable, orig_teardown: Callable
    ):
        '''
        Instantiate the resources required by the specific testcase (i.e either app/client or both)
        Then call `_handle_try_finally_around_internal_call`, passing in the necessary parameters depending
        on the test case

        This is to be implemented by the actual testcase class (down below)
        Since each testcase instantiates different resources differently
        '''
        raise NotImplementedError

    def _handle_try_finally_around_internal_call(
        self,
        internal_method: Callable,
        orig_test: Callable,
        orig_setup: Callable,
        orig_teardown: Callable,
        *test_args,
        create_app_result: Union[Flask, Iterator[Flask], None] = None
    ):
        '''
        Override the test method and setUp, tearDown by preparing them with *test_args
        content of *test_args varies by testcase and should be passed by `_handle_resource_instantiation_and_internal_call`
        
        Then call the internal method (prepared `super().run`/`super().debug`)

        Then cleanup
        '''
        try:
            '''
            Override the method to create a partially built method - pass in the app/client or both (i.e *test_args)

            `super().run`/`super().debug` can now call this test method without passing anything and it'll all work out

            NOTE: if `internal_method` references `super().run` - it is already prepared with the required `TestResult` object
            so it can be called directly, same as `super().debug` - which does not require any parameters and hence can be
            called directly too
            The preparation is done in `_TestCaseImpl::run` (up above)
            '''
            setattr(self, self._testMethodName, functools.partial(orig_test, *test_args))
            # Also override the set up and tear down methods similarly
            self.setUp = functools.partial(orig_setup, *test_args)
            self.tearDown = functools.partial(orig_teardown, *test_args)
            # Call the actual test
            return internal_method()
        finally:
            if hasattr(self, '_teardown_create_app_result'):
                # If the object has _teardown_create_app_result method - it's a AppTestCase/AppClientTestCase
                # In which case, handle tearing down the result gotten from `create_app` (should be passed as an arg)
                getattr(self, '_teardown_create_app_result')(create_app_result)

            # Restore the original methods
            setattr(self, self._testMethodName, orig_test)
            self.setUp = orig_setup
            self.tearDown = orig_teardown


class ClientTestCase(_TestCaseImpl):
    '''
    Test your flask web app using a FlaskClient object

    A FlaskClient object is created using the given app object for **each**
    test set (i.e setUp, test method, tearDown) - this same object is then passed
    to the setUp, test method and tearDown
    The user can use this FlaskClient object to test api calls

    A built and instantiated Flask app object should
    be assigned to the `app` property

    Can be used with unittest.TestSuite
    '''
    app: Union[Flask, None] = None
    # Whether or not to use cookies in test client
    test_client_use_cookies: bool = True
    # kwargs to pass to test_client function
    test_client_kwargs: Dict = {}

    def __init__(self, methodName='runTest'):
        # Verify self.app was provided
        if not self.app:
            raise NotImplementedError('property `app` must be assigned in ClientTestCase')
        # Call the original __init__
        super().__init__(methodName)

    def setUp(self, client: FlaskClient) -> None:
        '''
        Set up to do before running each test
        '''
        pass

    def tearDown(self, client: FlaskClient) -> None:
        '''
        Cleanup to do after running each test
        '''
        pass

    def _handle_resource_instantiation_and_internal_call(
        self, internal_method: Callable, orig_test: Callable, orig_setup: Callable, orig_teardown: Callable
    ):
        '''
        ClientTestCase only needs to instantiate the client, in a with block
        Do just that, and pass the client to `_handle_try_finally_around_internal_call`, which will then
        pass it to the actual test method, setUp and tearDown
        '''
        with self.app.test_client(self.test_client_use_cookies, **self.test_client_kwargs) as client:
            return self._handle_try_finally_around_internal_call(
                internal_method, orig_test, orig_setup, orig_teardown, client
            )


class AppTestCase(_TestCaseImpl):
    '''
    Test your flask web app using a Flask object

    A Flask object is created by calling the create_app function for **each**
    test set (i.e setUp, test method, tearDown) - this same object is then passed
    to the setUp, test method and tearDown
    The user can use this Flask object to test the app

    The `create_app` function should create/configure/set up and return/yield
    a Flask app object

    Can be used with unittest.TestSuite
    '''
    def create_app(self) -> Union[Flask, Iterator[Flask]]:
        '''
        Should return/yield a built/configured Flask app object

        To be implemented by the user
        '''
        raise NotImplementedError

    def setUp(self, app: Flask) -> None:
        '''
        Set up to do before
        '''
        pass

    def tearDown(self, app: Flask) -> None:
        pass

    ### Private helper methods

    def _instantiate_app(self):
        if isgeneratorfunction(self.create_app):
            # create_app yields an iterator/generator
            res = self.create_app()
            return res, next(res)
        else:
            # create_app returns a regular object - hopefully a Flask object
            res = self.create_app()
            if not isinstance(res, Flask):
                raise TypeError(f'Expected create_app to return a Flask object - got {type(res)}')
            return res, res

    def _handle_resource_instantiation_and_internal_call(
        self, internal_method: Callable, orig_test: Callable, orig_setup: Callable, orig_teardown: Callable
    ):
        '''
        AppTestCase needs to instantiate the app
        It also has to handle generator/iterator cases, so use `_instantiate_app()`
        and pass the `res` to `_handle_try_finally_around_internal_call` as `create_app_result`
        which will then handle tearing it down if it was a generator/iterator
        
        Otherwise, the `app` needs to be passed to `_handle_try_finally_around_internal_call`, which will then
        pass it to the actual test method, setUp and tearDown
        '''
        res, app = self._instantiate_app()
        return self._handle_try_finally_around_internal_call(
            internal_method, orig_test, orig_setup, orig_teardown, app, create_app_result=res
        )

    def _teardown_create_app_result(self, res: Union[Flask, Iterator[Flask]]):
        # Tear down the result obtained by calling create_app
        # This is only here to handle when create_app returns a generator/iterator
        if isinstance(res, Flask):
            # create_app did not return a generator/iterator - nothing to clean up
            return
        else:
            # Teardown the generator/iterator
            try:
                next(res)
            except StopIteration:
                pass
            else:
                raise TypeError('Expected create_app to yield 1 value - got multiple')


class AppClientTestCase(AppTestCase):
    '''
    Test your flask web app using a Flask object **and** a FlaskClient object

    A Flask and a FlaskClient object are created by calling the create_app function, and
    the `test_client` method on the Flask object for **each**
    test set (i.e setUp, test method, tearDown) - these same objects are then passed
    to the setUp, test method and tearDown
    The user can use these Flask and FlaskClient objects to test the app

    The `create_app` function should create/configure/set up and return
    a Flask app object

    Can be used with unittest.TestSuite
    '''
    # Whether or not to use cookies in test client
    test_client_use_cookies: bool = True
    # kwargs to pass to test_client function
    test_client_kwargs: Dict = {}

    def setUp(self, app: Flask, client: FlaskClient) -> None:
        '''
        Set up to do before running each test
        '''
        pass

    def tearDown(self, app: Flask, client: FlaskClient) -> None:
        '''
        Cleanup to do after running each test
        '''
        pass

    def _handle_resource_instantiation_and_internal_call(
        self, internal_method: Callable, orig_test: Callable, orig_setup: Callable, orig_teardown: Callable
    ):
        '''
        AppClientTestCase needs to instantiate the app *and* the client
        It also has to handle generator/iterator cases, just like AppTestCase
        
        The `app` and `client` needs to be passed to `_handle_try_finally_around_internal_call`, which will then
        pass it to the actual test method, setUp and tearDown
        '''
        res, app = self._instantiate_app()
        with app.test_client(use_cookies=self.test_client_use_cookies, **self.test_client_kwargs) as client:
            return self._handle_try_finally_around_internal_call(
                internal_method, orig_test, orig_setup, orig_teardown, app, client, create_app_result=res
            )



import unittest
from typing import Union

from .suite import LiveTestSuite
from .utils import _partialclass

# A constant placeholder to signify default timeout parameter
_GLOBAL_DEFAULT_TIMEOUT = object()


class LiveTestLoader(unittest.TestLoader):
    suiteClass = LiveTestSuite

    def __init__(self, flask_app, timeout: Union[float, None] = _GLOBAL_DEFAULT_TIMEOUT):
        '''
        Partially construct the TestSuite class
        unittest.TestLoader works with unittest.TestSuite,
        which does not take an `flask_app` (also `timeout`) as a parameter during construction

        To avoid overriding every single method in LiveTestLoader just to pass in that
        extra `flask_app` (also `timeout`) parameter to `suiteClass` - turn suiteClass into a partially constructed LiveTestSuite
        This will make the construction process of LiveTestSuite (in later stages), exactly
        the same as unittest.TestSuite

        Basically, doing stuff like `self.suiteClass(map(testCaseClass, testCaseNames))`
        (something that is done in unittest.TestLoader) will work perfectly fine because the `flask_app` (also `timeout`)
        parameter has already been passed to `suiteClass` - but it hasn't been instantiated - it's still a class!
        '''

        self.suiteClass = _partialclass(
            LiveTestSuite, flask_app, timeout if timeout is not _GLOBAL_DEFAULT_TIMEOUT else None
        )
        super().__init__()


# A constant placeholder to signify default live test loader parameter
defaultLiveTestLoader = object()



import unittest
from typing import Union

from flask import Flask

from .loader import LiveTestLoader, defaultLiveTestLoader

# A constant placeholder to signify default timeout parameter
_GLOBAL_DEFAULT_TIMEOUT = object()

class LiveTestProgram(unittest.TestProgram):
    def __init__(self, flask_app: Flask, timeout: Union[float, None]=_GLOBAL_DEFAULT_TIMEOUT, module='__main__', defaultTest=None, argv=None,
                    testRunner=None, testLoader: Union[LiveTestLoader, None]=defaultLiveTestLoader,
                    exit=True, verbosity=1, failfast=None, catchbreak=None,
                    buffer=None, warnings=None, *, tb_locals=False):
        if testLoader is defaultLiveTestLoader:
            # Construct the default live test loader if no custom loader was provided
            testLoader = LiveTestLoader(flask_app, timeout if timeout is not _GLOBAL_DEFAULT_TIMEOUT else None)
        super().__init__(module, defaultTest, argv, testRunner, testLoader, exit, verbosity, failfast, catchbreak, buffer, warnings, tb_locals=tb_locals)

main_live = LiveTestProgram



from setuptools import setup

with open('README.md', 'r') as readmefile:
    readme = readmefile.read()

setup(
    name='flask-unittest',
    version='0.1.2',
    url='https://github.com/TotallyNotChase/flask-unittest',
    license='MIT',
    author='TotallyNotChase',
    author_email='totallynotchase42@gmail.com',
    description='Unit testing flask applications made easy!',
    long_description=readme,
    long_description_content_type='text/markdown',
    packages=['flask_unittest'],
    test_suite="tests.normalsuite",
    platforms='any',
    install_requires=['Flask>=1.1.0'],
    tests_require=['selenium', 'beautifulsoup4'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Intended Audience :: Developers',
    ],
    python_requires='>=3.6',
)



import functools


def _partialclass(cls, *args, **kwds):
    '''
    Return a partially constructed class from given class

    Essentially the same `functools.partial` but for classes
    
    from: https://stackoverflow.com/a/38911383/10305477
    '''
    class NewCls(cls):
        __init__ = functools.partialmethod(cls.__init__, *args, **kwds)

    return NewCls



import unittest
import threading
import socket
from typing import Union, Iterator, Iterable

from flask import Flask

from .case import LiveTestCase

_TestType = Union[LiveTestCase, unittest.TestSuite]

# A constant placeholder to signify default timeout parameter
_GLOBAL_DEFAULT_TIMEOUT = object()

# Store localhost as constant
_LOCALHOST = '127.0.0.1'


class LiveTestSuite(unittest.TestSuite):
    # Handle for the flask server
    _thread: Union[threading.Thread, None] = None

    def __init__(
        self, flask_app: Flask, timeout: Union[float, None] = _GLOBAL_DEFAULT_TIMEOUT, tests: Iterable[_TestType] = ()
    ):
        self._app = flask_app
        self._timeout = timeout if timeout is not _GLOBAL_DEFAULT_TIMEOUT else None
        self._host: str = flask_app.config.get('HOST', '127.0.0.1')
        self._port: int = flask_app.config.get('PORT', 5000)
        super().__init__(tests)

    def run(self, result, debug=False):
        self._setup_server()
        self._setup_testcases()
        return super().run(result, debug)

    ### Private helper methods

    def _setup_testcases(self):
        # Set up required properties in all testcases in current testsuite
        def _inject_properties(testcase: LiveTestCase):
            # Inject the required properties into the given test case
            testcase.server_url = f'http://127.0.0.1:{self._port}'
            testcase.app = self._app

        for test in self:
            if self._isnotsuite(test):
                _inject_properties(test)
            else:
                # Current element is an entire suite - iterate through it and add properties to each test
                for inner_test in test:
                    _inject_properties(inner_test)

    def _setup_server(self):
        # Spawn the flask server as a separate process
        self._thread = threading.Thread(target=self._app.run, kwargs={'host': self._host, 'port': self._port, 'use_reloader': False})
        self._thread.setDaemon(True)
        self._thread.start()
        # Wait for the server to start responding, until a specific timeout
        sckt = socket.create_connection((_LOCALHOST, self._port), timeout=self._timeout or None)
        sckt.close()

    def _isnotsuite(self, test):
        '''
        A crude way to tell apart testcases and suites with duck-typing

        Taken from unittest.TestSuite internals
        '''
        try:
            iter(test)
        except TypeError:
            return True
        return False

    ### Override the type hints of some derived functions

    def addTest(self, test: _TestType):
        super().addTest(test)

    def addTests(self, tests: Iterable[_TestType]):
        super().addTests(tests)

    def __iter__(self) -> Iterator[_TestType]:
        return super().__iter__()
